# -*- coding=Latin1 -*-
class context():
    # basic
    arch='amd64' # x86
    endian='little' # big
    log_level=""
    timeout=512
    tick=0.0625
    newline='\r\n'
    pie=None

    # input output
    noout=None
    nocolor=None

    # debug
    dbginit="" # debugger init command
    windbg_path=None # set windbg.exe path 



import collections
import logging
import math
import operator
import os
import re
import requests
import socket
import signal
import string
import struct
import subprocess
import sys
import tempfile
import threading
import time

from q4nwinlib.interactive.winpwn import process, remote
from q4nwinlib.misc import p32, p16, p8, p64, u8, u16, u32, u64, pause, PIE, NOPIE, sleep, Latin1_encode, Latin1_decode, color, hexdump, flat, fit, encode_hex, decode_hex
from q4nwinlib.misc.log import Log
from q4nwinlib.context import context
from q4nwinlib.dbg import windbg, windbgx
from q4nwinlib.asm import asm, disasm

__all__ = [x for x in tuple(globals()) if x != '__name__']



# -*- coding=Latin1 -*-
# ctypes:
#   create_string_buffer
#   create_unicode_buffer
# windll.kernel32.CreateProcessA
# windll.kernel32.CreatePipe
# windll.kernel32.PeekNamedPipe
# windll.kernel32.ReadFile
# windll.kernel32.WriteFile
# windll.kernel32.SetHandleInformation
# windll.kernel32.GetStdHandle
# windll.kernel32.CloseHandle

import os
import sys
import time

from ctypes import windll,byref,sizeof,wintypes,create_string_buffer,GetLastError,c_size_t
from ctypes.wintypes import HANDLE,LPVOID,LPSTR,DWORD,WORD,BOOL,BYTE
from ctypes import POINTER,Structure

from q4nwinlib.context import context
from q4nwinlib.misc import Latin1_encode,Latin1_decode,color,showbanner

# some var to CreatePipe or CreateProcessA
HANDLE_FLAG_INHERIT=1
STARTF_USESTDHANDLES=256
STILL_ACTIVE=259
CREATE_SUSPENDED=0x4
PROCESS_VM_READ=0x10
PROCESS_VM_WRITE=0x20
PROCESS_VM_OPERATION=0x8
PAGE_READWRITE=0x4

# for createPipe
class SECURITY_ATTRIBUTES(Structure):
    _fields_= [
        ('nLength',DWORD),
        ('lpSecurityDescriptor',LPVOID),
        ('bInheritHandle',BOOL),
    ]
# for createProcess
class PROCESS_INFORMATION(Structure):
    _fields_= [
        ('hProcess', HANDLE),
        ('hThread', HANDLE),
        ('dwProcessId',DWORD),
        ('dwThreadId',DWORD),
    ]
# for createProcess
class STARTUPINFO(Structure):
    _fields_=[
        ('cb',DWORD),
        ('lpReserved',LPSTR),
        ('lpDesktop',LPSTR),
        ('lpTitle',LPSTR),
        ('dwX',DWORD),
        ('dwY',DWORD),
        ('dwXSize',DWORD),
        ('dwYSize',DWORD),
        ('dwXCountChars',DWORD),
        ('dwYCountChars',DWORD),
        ('dwFillAttribute',DWORD),
        ('dwFlags',DWORD),
        ('wShowWindow',WORD),
        ('cbReserved2',WORD),
        ('lpReserved2',POINTER(BYTE)),
        ('hStdInput',HANDLE),
        ('hStdOutput',HANDLE),
        ('hStdError',HANDLE),
    ]

# class winApi():
    # @classmethod
    # def VirtualAllocEx(clx):
    #     pass
    # @classmethod
    # def VirtualFreeEx(clx,addr,size):
    #     pass
    # @classmethod
    # def GetModuleHandle(clx):
    #     pass
    # @classmethod
    # def VirtualProtectEx(clx,hProcess,addr,size,protect,oldprotect):
    #     return windll.kernel32.VirtualProtectEx(hProcess,)
    # @classmethod
    # def ReadProcessMemory(clx,hProcess,addr,n):
    #     beenRead=wintypes.DWORD()
    #     buf=create_string_buffer(n)
    #     x=windll.kernel32.ReadProcessMemory(self.hProcess,addr,buf,n,byref(beenRead))
    #     if x==0:
    #         raise(EOFError())
    #     return Latin1_decode(buf.raw)
    # @classmethod
    # def WriteProcessMemory(clx,hProcess,addr,buf):
    #     n=len(buf)
    #     written=wintypes.DWORD()
    #     x=windll.kernel32.WriteProcessMemory(self.hProcess,addr,buf,n,byref(written))
    #     if x==0:
    #         raise(EOFError())
    #     return written.value
    # @classmethod
    # def ResumeThread(clx,hThread):
    #     return windll.kernel32.ResumeThread(hThread)

class winPipe():
    def __init__(self,bInheritHandle = 1):
        self.timeout=context.timeout
        self.hReadPipe,self.hWritePipe,self.child_hReadPipe,self.child_hWritePipe=self.create(bInheritHandle=bInheritHandle)

    def create(self,bInheritHandle = 1):
        # set attr
        attr=SECURITY_ATTRIBUTES()
        attr.lpSecurityDescriptor=0
        attr.bInheritHandle = bInheritHandle
        attr.nLength = sizeof(attr)

        hReadPipe=wintypes.HANDLE()      # father read, child write
        hWritePipe=wintypes.HANDLE()     # father write, child read
        child_hReadPipe=wintypes.HANDLE()
        child_hWritePipe=wintypes.HANDLE()

        rs1=windll.kernel32.CreatePipe(byref(hReadPipe),byref(child_hWritePipe),byref(attr),0)
        rs2=windll.kernel32.CreatePipe(byref(child_hReadPipe),byref(hWritePipe),byref(attr),0)
        
        rs3=windll.kernel32.SetHandleInformation(hReadPipe.value,HANDLE_FLAG_INHERIT,0)
        rs4=windll.kernel32.SetHandleInformation(hWritePipe.value,HANDLE_FLAG_INHERIT,0)

        if(rs1 and rs2 and rs3 and rs4):
            return (hReadPipe.value,hWritePipe.value,child_hReadPipe.value,child_hWritePipe.value)
        else:
            raise(EOFError(color("[-]: Create Pipe error",'red')))

    def read(self,n,timeout=None):
        def count():
            byteAvail=wintypes.DWORD()
            x=windll.kernel32.PeekNamedPipe(self.hReadPipe,0,0,0,byref(byteAvail),0)
            return byteAvail.value
        if timeout is None:
            if self.timeout:
                timeout=self.timeout
            else:
                timeout=context.timeout
        x_time=0
        if count()<n:
            while(x_time<timeout and count()<n):
                time.sleep(context.tick)
                x_time+=context.tick
        cn=min(count(),n)
        beenRead=wintypes.DWORD()
        buf=create_string_buffer(cn)
        if cn>0:
            windll.kernel32.ReadFile(self.hReadPipe,buf,cn,byref(beenRead),None)
        return Latin1_decode(buf.raw)

    def write(self,buf=''):
        buf=Latin1_encode(buf)
        length=len(buf)
        written=wintypes.DWORD()
        x=windll.kernel32.WriteFile(self.hWritePipe,buf,length,byref(written),None)
        if x==0:
            raise(EOFError())
        return written.value
    
    def getHandle(self):
        return (self.hReadPipe,self.hWritePipe,self.child_hReadPipe,self.child_hWritePipe)

    def close(self):
        # windll.kernel32.CloseHandle(self.child_hReadPipe)
        # windll.kernel32.CloseHandle(self.child_hWritePipe)
        windll.kernel32.CloseHandle(self.hReadPipe)
        windll.kernel32.CloseHandle(self.hWritePipe)

class winProcess(object):
    def __init__(self,argv,cwd=None,flags=0):
        self.pipe=winPipe()
        self.hReadPipe,self.hWritePipe,self.child_hReadPipe,self.child_hWritePipe=self.pipe.getHandle()
        self.pid=0
        self.phandle=0
        self.tid=0
        self.thandle=0
        self.create(argv,cwd,flags)
    def create(self,argv,cwd=None,flags=None):
        lpCurrentDirectory=cwd
        lpEnvironment=None
        dwCreationFlags=flags
        bInheritHandles=True
        lpProcessAttributes=None
        lpThreadAttributes=None

        lpProcessInformation = PROCESS_INFORMATION()

        StartupInfo = STARTUPINFO()
        StartupInfo.cb = sizeof(StartupInfo)

        StartupInfo.dwFlags = STARTF_USESTDHANDLES
        StartupInfo.hStdInput = self.child_hReadPipe
        StartupInfo.hStdOutput =self.child_hWritePipe
        StartupInfo.hStdError = self.child_hWritePipe
        
        lpStartupInfo = byref(StartupInfo)
        
        lpCommandLine =  None
        lpApplicationName = None

        if not isinstance(argv,list):
            lpApplicationName = Latin1_encode(argv)
        else:
            lpCommandLine = Latin1_encode((" ".join([str(a) for a in argv])))
        try: 
            bs=windll.kernel32.CreateProcessA(
                lpApplicationName,          
                lpCommandLine,              
                lpProcessAttributes,        
                lpThreadAttributes,         
                bInheritHandles,            
                dwCreationFlags,            
                lpEnvironment,              
                lpCurrentDirectory,         
                byref(StartupInfo),         
                byref(lpProcessInformation)
            )
            self.pid=lpProcessInformation.dwProcessId
            self.phandle=lpProcessInformation.hProcess
            self.tid=lpProcessInformation.dwThreadId
            self.thandle=lpProcessInformation.hThread
            showbanner('Create process success #pid 0x{:x}'.format(self.pid))
        except:
            raise(EOFError(color("[-]: Create process error",'red')))

    def read(self,n,timeout=None):
        return self.pipe.read(n,timeout=timeout)
    def write(self,buf):
        return self.pipe.write(buf)
    def is_exit(self):
        x=wintypes.DWORD()
        n=windll.kernel32.GetExitCodeProcess(self.phandle,byref(x))
        if n!=0 and x.value==STILL_ACTIVE:
            return False
        return True
    def close(self):           # need to kill process ..............
        self.pipe.close()
        windll.kernel32.TerminateProcess(self.phandle,1)
    def readm(self,addr,n):
        addr=c_size_t(addr)
        handle=windll.kernel32.OpenProcess(PROCESS_VM_READ|PROCESS_VM_WRITE|PROCESS_VM_OPERATION,0,self.pid)
        oldprotect=wintypes.DWORD()
        x=windll.kernel32.VirtualProtectEx(handle,addr,n,PAGE_READWRITE,byref(oldprotect))
        
        buf=create_string_buffer(n)
        
        x=windll.kernel32.ReadProcessMemory(handle,addr,buf,n,0)
        if x==0:
            raise(MemoryError)

        windll.kernel32.VirtualProtectEx(handle,addr,n,oldprotect.value,0)
        windll.kernel32.CloseHandle(handle)

        return Latin1_decode(buf.raw)

    def writem(self,addr,buf):
        buf=Latin1_encode(buf)
        addr=c_size_t(addr)
        n=len(buf)
        # print(n)
        handle=windll.kernel32.OpenProcess(PROCESS_VM_READ|PROCESS_VM_WRITE|PROCESS_VM_OPERATION,0,self.pid)
        
        oldprotect=wintypes.DWORD()
        x=windll.kernel32.VirtualProtectEx(handle,addr,n,PAGE_READWRITE,byref(oldprotect))
        written=c_size_t(0)
        x=windll.kernel32.WriteProcessMemory(handle,addr,buf,n,byref(written))
        if x==0:
            raise(MemoryError)
        
        windll.kernel32.VirtualProtectEx(handle,addr,n,oldprotect.value,0)
        windll.kernel32.CloseHandle(handle)

        return written.value


from q4nwinlib.context import context
from q4nwinlib.misc import Latin1_encode,Latin1_decode

def disasm(machine_code,addr=0,arch=None):
    import capstone
    machine_code=Latin1_encode(machine_code)
    if arch is None:
        arch=context.arch
    if arch=='i386':
        disasmer=capstone.Cs(capstone.CS_ARCH_X86, capstone.CS_MODE_32)
    elif arch=="amd64":
        disasmer=capstone.Cs(capstone.CS_ARCH_X86, capstone.CS_MODE_64)
    l=""
    for i in disasmer.disasm(machine_code,addr):
        l+="{:8s} {};\n".format(i.mnemonic,i.op_str)
    return Latin1_decode(Latin1_encode(l.strip('\n')))

def asm(asm_code,addr=0,arch=None):
    import keystone
    asm_code=Latin1_encode(asm_code)
    if arch is None:
        arch=context.arch
    if arch=='i386':
        asmer=keystone.Ks(keystone.KS_ARCH_X86,keystone.KS_MODE_32)
    elif arch=="amd64":
        asmer=keystone.Ks(keystone.KS_ARCH_X86,keystone.KS_MODE_64)
    l=""
    for i in asmer.asm(asm_code,addr)[0]:
        l+=chr(i)
    return Latin1_decode(Latin1_encode(l.strip('\n')))





# -*- coding=Latin1 -*-
import platform

import threading
import sys
import socket
import time
import os

from q4nwinlib.interactive.win import winProcess
from q4nwinlib.context import context
from q4nwinlib.misc import Latin1_encode,Latin1_decode,NOPIE,PIE,color,showbanner,showbuf,pause
from q4nwinlib.dbg import windbgx

class tube(object):
    def __init__(self):
        self._timeout=context.timeout
        self.debugger=None
        
    def read(self,n,timeout=None,interactive=False):
        pass
    def write(self,buf):
        pass
    def is_exit(self,buf):
        pass
    @property
    def timeout(self):
        return self._timeout
    @timeout.setter
    def timeout(self,timeout):
        pass             
    def send(self,buf):
        showbanner('Send')
        rs=self.write(buf)
        showbuf(buf)
        return rs
    
    def sendline(self,buf,newline=None):
        if newline is None:
            newline=context.newline
        return self.send(buf+newline)

    def recv(self,n,timeout=None):
        # try to read n bytes, no exception
        showbanner('Recv')
        buf=self.read(n, timeout)
        showbuf(buf)
        return buf

    def recvn(self,n,timeout=None):
        # must recv n bytes within timeout
        showbanner("Recv")
        buf = self.read(n, timeout)
        if len(buf) != n:
            raise(EOFError(color("[-]: Timeout when use recvn",'red')))
        showbuf(buf)
        return buf
    
    def recvuntil(self,delim,drop = True,timeout=None):
        showbanner("Recv")
        if timeout is None:
            if self.timeout:
                timeout=self.timeout
            else:
                timeout=context.timeout
        buf = ''
        st=time.time()
        xt=st
        while (buf[-len(delim):]!=delim):
            buf += self.read(1, timeout=timeout-(xt-st))
            if self.debugger is None:
                xt=time.time()
                if (xt-st)>=timeout:
                    break
        if not buf.endswith(delim):
            raise(EOFError(color("[-]: Recvuntil error",'red')))
        if drop:
            buf = buf[:-len(delim)]
        showbuf(buf)
        return buf

    def recvline(self,drop = True,timeout=None,newline=None):
        if newline is None:
            newline=context.newline
        return self.recvuntil(newline,drop=drop)

    def recvall(self,timeout=None):
        showbanner('Recv')
        buf=self.read(0x100000, timeout)
        showbuf(buf)
        return buf   

    # based on read/write
    def interactive(self):
        # it exited, contrl+C, timeout
        showbanner('Interacting',is_noout=False)
        go = threading.Event()
        go.clear()
        def recv_thread():
            try:                  
                while not go.is_set():
                    buf = self.read(0x10000,0.125,interactive=True)
                    if buf:
                        showbuf(buf,is_noout=False)
                        showbanner('Interacting',is_noout=False)
                    go.wait(0.2)
            except KeyboardInterrupt:
                go.set()
                print(color('[-]: Exited','red'))
        t = threading.Thread(target = recv_thread)
        t.daemon = True
        t.start()

        try:
            while not go.is_set():
                go.wait(0.2)
                try:
                    if self.is_exit():
                        time.sleep(0.2) # wait for time to read output
                    buf = sys.stdin.readline()
                    if buf:
                        self.write(buf)           # remote.write() may cause exception
                except:
                    go.set()
                    print(color('[-]: Exited','red'))
                    break
        except KeyboardInterrupt: # control+C
            go.set()
            print(color('[-]: Exited','red'))
        while t.is_alive():
            t.join(timeout = 0.1)
    
    def sendlineafter(self, a,b,drop=False,timeout=None):
        self.recvuntil(a,drop=drop,timeout=timeout)
        self.sendline(b)
    def sendafter(self, a,b,drop=False,timeout=None):
        self.recvuntil(a,drop=drop,timeout=timeout)
        self.send(b)
    
    # alias
    ia = interactive
    sd = send
    sl = sendline
    rn = recvn
    rv = recv
    ru = recvuntil
    ra = recvall
    rl = recvline
    sa = sendafter
    sla = sendlineafter


class remote(tube):
    def __init__(self, ip, port, family = socket.AF_INET, socktype = socket.SOCK_STREAM):
        tube.__init__(self)
        self.sock = socket.socket(family, socktype)
        # self.ip = ip
        # self.port = port
        self._is_exit=False
        try:
            showbanner("Connecting to ({},{})".format(ip,port))
            self.sock.settimeout(self.timeout)
            self.sock.connect((ip, port))
        except:
            raise(EOFError(color("[-]: Connect to ({},{}) failed".format(ip,port),'red')))
    def read(self,n,timeout=None,interactive=False):
        if timeout is not None:
            self.sock.settimeout(timeout)
        buf=b''
        try:
            buf=self.sock.recv(n) # ignore timeout error
        except KeyboardInterrupt:
            self.close()
            raise(EOFError(color("[-]: Exited by CTRL+C",'red')))
        except:
            pass
        self.sock.settimeout(self.timeout)
        return Latin1_decode(buf)
    def write(self,buf):
        return self.sock.send(Latin1_encode(buf))
    def close(self):
        self.sock.close()
        self._is_exit=True
    def is_exit(self):
        if self._is_exit:
            return True
        return False
    @tube.timeout.setter
    def timeout(self,timeout):
        self._timeout=timeout
        self.sock.settimeout(self._timeout)
    
    def dbg(self, script = ''):
        pass

class process(tube):
    def __init__(self,argv,cwd=None,flags=None):
        tube.__init__(self)
        # en/disable PIE, need: pip install pefile
        if context.pie is not None:
            fpath=""
            if not isinstance(argv,list):
                fpath=argv
            else:
                fpath=argv[0]
            if context.pie:
                PIE(fpath)
            else:
                NOPIE(fpath)
        self.Process=winProcess(argv,cwd,flags)
        self.pid=self.Process.pid
    def read(self,n,timeout=None,interactive=False):
        buf=''
        try:
            if self.debugger is not None and interactive is False:
                while(len(buf)!=n):
                    buf+=self.Process.read(n-len(buf),timeout)
            else:
                buf=self.Process.read(n, timeout)
        except KeyboardInterrupt:
            self.close()
            raise(EOFError(color("[-]: Exited by CTRL+C",'red')))       
        return buf
    def write(self, buf):
        return self.Process.write(buf)
    def readm(self,addr,n):
        return self.Process.readm(addr,n)
    def writem(self,addr,con):
        return self.Process.writem(addr,con)
    def close(self):
        self.Process.close()        # need to kill process
    def is_exit(self):
        return self.Process.is_exit()
    @tube.timeout.setter
    def timeout(self,timeout):
        self._timeout=timeout
        self.Process.pipe.timeout=timeout

    def dbg(self, script = ''):
        windbgx.attach(self,script)
        pause() # waiting dbgger


from __future__ import absolute_import


from q4nwinlib.misc import log

__all__ = ['log']

# -*- coding=Latin1 -*-
# import os
import subprocess
import struct
import time
import sys
import os
import six

from q4nwinlib.context import context

def NOPIE(fpath=""):
    import pefile
    pe_fp=pefile.PE(fpath)
    pe_fp.OPTIONAL_HEADER.DllCharacteristics &= \
        ~pefile.DLL_CHARACTERISTICS["IMAGE_DLLCHARACTERISTICS_DYNAMIC_BASE"]
    pe_fp.OPTIONAL_HEADER.CheckSum = pe_fp.generate_checksum()
    pe_fp.write(fpath)
def PIE(fpath=""):
    import pefile
    pe_fp=pefile.PE(fpath)
    pe_fp.OPTIONAL_HEADER.DllCharacteristics |= \
        pefile.DLL_CHARACTERISTICS["IMAGE_DLLCHARACTERISTICS_DYNAMIC_BASE"]
    pe_fp.OPTIONAL_HEADER.CheckSum = pe_fp.generate_checksum()
    pe_fp.write(fpath)

def pause(string=None):
    print(color("\n[=]: pausing",'purple'))
    if string is not None:
        print(color(string,'purple'))
    sys.stdin.readline()
def sleep(n):
    time.sleep(n)

def p64(i):
    endian = None
    if context.endian == 'little':
        endian = '<'
    else:
        endian = '>'
    l=struct.pack(endian+'Q', i)
    if sys.version_info[0]==3:
        return Latin1_decode(l)
    return l
def u64(s):
    endian = None
    if context.endian == 'little':
        endian = '<'
    else:
        endian = '>'
    return struct.unpack(endian+'Q', Latin1_encode(s))[0]
def p32(i):
    endian = None
    if context.endian == 'little':
        endian = '<'
    else:
        endian = '>'
    l=struct.pack(endian+'I', i)
    if sys.version_info[0]==3:
        return Latin1_decode(l)
    return l
def u32(s):
    endian = None
    if context.endian == 'little':
        endian = '<'
    else:
        endian = '>'
    return struct.unpack(endian+'I', Latin1_encode(s))[0]
def p16(i):
    endian = None
    if context.endian == 'little':
        endian = '<'
    else:
        endian = '>'
    l=struct.pack(endian+'H', i)
    if sys.version_info[0]==3:
        return Latin1_decode(l)
    return l
def u16(s):
    endian = None
    if context.endian == 'little':
        endian = '<'
    else:
        endian = '>'
    return struct.unpack(endian+'H', Latin1_encode(s))[0]
def p8(i):
    endian = None
    if context.endian == 'little':
        endian = '<'
    else:
        endian = '>'
    l=struct.pack(endian+'B', i)
    if sys.version_info[0]==3:
        return Latin1_decode(l)
    return l
def u8(s):
    endian = None
    if context.endian == 'little':
        endian = '<'
    else:
        endian = '>'
    return struct.unpack(endian+'B', Latin1_encode(s))[0]

def Latin1_encode(string): # -> bytes
    ''' Latin1_encode(string: str): -> bytes '''
    if sys.version_info[0]==3:
        return bytes(string,"Latin1")
    return bytes(string)
def Latin1_decode(string): # -> str
    ''' Latin1_encode(string: bytes): -> str '''
    if sys.version_info[0]==3:
        return str(string,'Latin1')
    return string

def encode_hex(string):
    ''' encode_hex(string: str) -> str
>>> encode_hex('aaabcccd')
'6161616263636364'
>>>
    '''
    if sys.version_info[0]==3:
        return hex(int.from_bytes(Latin1_encode(string),'big'))[2:]
    return string.encode('hex')

def decode_hex(string):
    ''' decode_hex(string: str) -> str
>>> decode_hex('11aabbccdd')
'\x11....'
>>> bytes(decode_hex('11aabbccdd'),"Latin1")
b'\x11\xaa\xbb\xcc\xdd'
>>>
    '''
    if sys.version_info[0]==3:
        return Latin1_decode(bytes.fromhex(string))
    return string.decode('hex')

def color(content,color='purple'):
    if context.nocolor:
        return content
    c = {
        "black": 30,
        "red": 31,
        "green": 32,
        "yellow": 33,
        "blue": 34,
        "purple": 35,
        "cyan": 36,
        "white": 37,
    }
    return "\033[0;{}m{}\033[0m".format(c.get(color), content)

def hexdump(src,length=16,all=True):
    FILTER = ''.join([(len(repr(chr(x))) == 3) and chr(x) or '.' for x in range(256)])
    lines = []
    for c in range(0, len(src), length):
        chars = src[c:c+length]
        hex=''
        printable=''
        for i in range(len(chars)):
            chex="%02x " % ord(chars[i])
            pchar=("%s" % ((ord(chars[i]) <= 127 and FILTER[ord(chars[i])]) or '.'))
            if (i%4)==0:
                chex=' '+chex
                pchar=color('|','red')+pchar
            hex+=chex
            printable+=pchar
        lines.append(color(
            "{:04x}".format(c)) +
            "  {}  {}\n".format(color(hex.ljust(52,' '),'yellow'),printable)
        )
    if not all:
        if len(lines)>=0x20:
            lines=lines[0:8]+['......\n']+lines[-8:]
    print(''.join(lines).strip())
    return ''.join(lines).strip()
    
def showbanner(markstr,colorstr='green',typestr='[+]',is_noout=None):
    if is_noout is None:
        is_noout=context.noout
    if not is_noout:
        print(color('\n'+typestr+': '+markstr,colorstr))
def showbuf(buf,is_noout=None):
    if is_noout is None:
        is_noout=context.noout
    if not is_noout:
        if context.log_level=='debug':
            hexdump(buf)
        if buf.endswith(context.newline):
            os.write(sys.stdout.fileno(), Latin1_encode(buf))
        else:
            os.write(sys.stdout.fileno(), Latin1_encode(buf+'\n'))

def packer(value): # -> str
    if context.arch == "amd64":
        return p64(value)
    elif context.arch == "x86":
        return p32(value)
    else:
        raise Exception(color("no such context.arch",'red'))

def _flat(arg): # -> bytes
    if isinstance(arg, (list, tuple)):
        val = b''
        for it in arg:
            val += _flat(it)
    elif isinstance(arg, bytes):
        val = arg
    elif isinstance(arg, six.text_type):
        val = arg.encode('utf8')
    elif isinstance(arg, six.integer_types):
        val = Latin1_encode(packer(arg))
    elif isinstance(arg, bytearray):
        val = bytes(arg)    
    else:
        raise Exception(color("flat/fit doesn't support such type: "+str(type(arg)),'red'))
    return val
def flat(*args): # -> bytes
    # Flat simplified version
    result = b''
    for item in args:
        result += _flat(item)
    return result

fit = flat



# -*- coding=Latin1 -*-
import tempfile
import os
import sys
import subprocess

from q4nwinlib.context import context
from q4nwinlib.misc import showbanner,Latin1_encode,sleep,pause,color

class windbg():
    @classmethod
    def attach(clx,target,script="",sysroot=None):
        
        showbanner('attaching','purple','[=]')
        if context.windbg_path is None:
            raise Exception(color("can't find windbg.exe, plz set context.windbg_path",'red'))
        else:
            windbgPath=context.windbg
        load_windbg=[windbgPath,'-p']
        if isinstance(target,int):
            load_windbg.append(str(target))
        else:
            load_windbg.append(str(target.pid))

        script=context.dbginit+'\n'+script+'\n'
        tmp=tempfile.NamedTemporaryFile(prefix = 'winpwn_', suffix = '.dbg',delete=False)
        tmp.write(Latin1_encode(script))
        tmp.flush()
        tmp.close()
        load_windbg += ['-c']             # exec command
        load_windbg+=['$$><{}'.format(tmp.name)+';.shell -x del {}'.format(tmp.name)]
        # print('script:',script)
        # print('load:',load_windbg)
        ter=subprocess.Popen(load_windbg)
        while(os.path.exists(tmp.name)):    # wait_for_debugger
            pass
        target.debugger=ter
        return ter.pid

class windbgx():
    @classmethod
    def attach(clx,target,script="",sysroot=None):
        showbanner('attaching','purple','[=]')
        windbgxPath='windbgx'
        load_windbg=[windbgxPath,'-p']

        if isinstance(target,int):
            load_windbg.append(str(target)) # target is pid
        else:
            load_windbg.append(str(target.pid))
        script=context.dbginit+'\n'+script+'\n'

        tmp=tempfile.NamedTemporaryFile(prefix = 'winpwn_', suffix = '.dbg',delete=False)
        tmp.write(Latin1_encode(script))
        tmp.flush()
        tmp.close()
        load_windbg += ['-c']             # exec command
        load_windbg+=['"$$><{}'.format(tmp.name)+';.shell -x del {}"'.format(tmp.name)]
        # print('script:',script)
        # print('load:',load_windbg)
        ter=subprocess.Popen(' '.join(load_windbg))
        while(os.path.exists(tmp.name)):    # wait_for_debugger
            pass
        target.debugger=ter
        # mark('attached')
        return ter.pid




# -*- coding:utf-8 -*-
from __future__ import absolute_import

import importlib
__all__ = [
    'misc',
    'interactive',
    'context',
    'dbg',
    'asm'
]

for module in __all__:
    importlib.import_module('.%s' % module, 'q4nwinlib')